import os
os.environ.update({"KINETO_LOG_LEVEL": "99"})

from collections import defaultdict
from operator import itemgetter
from warnings import warn

from torch import Tensor
from torch.nn import Module
from torch.autograd import Function
from torch.utils._pytree import tree_map
from torch.utils._python_dispatch import TorchDispatchMode

from torch_operation_counter.counters import operations_mapping
from torch_operation_counter.ignore_modules import aten_ignore
from torch_operation_counter.utils import normalize_tuple

MAIN_NAME = "__main__"


class OperationsCounterMode(TorchDispatchMode):
    """
    A class to count the number of operations in a PyTorch model. This class inherits from TorchDispatchMode,
    which is a context manager for registering hooks for forward and backward methods of PyTorch modules.

    Attributes:
    -----------
    ops_count: dict
        A dictionary that counts the number of operations for each function.
    total_main_operation: int
        The total number of operations for the main function.
    """

    def __init__(self, module: Module = None):
        """
        :param module: nn.Module, module to be registered for counting operations.
        """
        super().__init__()
        self.operations_count = defaultdict(lambda: defaultdict(int))
        self.total_main_operation = None
        self.parents = [MAIN_NAME]
        if module is not None:
            for name, module in dict(module.named_children()).items():
                module.register_forward_pre_hook(self.enter_module(name))
                module.register_forward_hook(self.exit_module(name))

    def enter_module(self, name):
        """
        Registers a hook for the forward pre-hook of a module.

        :param name: The name of the PyTorch module.
        :return:The forward pre-hook function.
        """

        def f(module, x):
            self.parents.append(name)
            x = normalize_tuple(x)
            x = self.create_backwards_pop(name)(*x)
            return x

        return f

    def exit_module(self, name):
        """
        Registers a hook for the forward hook of a PyTorch module.

        :param name: The name of the module.
        :return: The forward hook function.
        """

        def f(module, x, y):
            assert self.parents[-1] == name
            self.parents.pop()
            y = normalize_tuple(y)
            return self.create_backwards_push(name)(*y)

        return f

    def create_backwards_push(self, name):
        """
        Creates a function for the backwards push.

        :param name: The name of the PyTorch module.
        :return: The function for the backwards push.
        """

        class PushState(Function):
            @staticmethod
            def forward(ctx, *args):
                args = tree_map(lambda x: x.clone() if isinstance(x, Tensor) else x, args)
                if len(args) == 1:
                    return args[0]
                return args

            @staticmethod
            def backward(ctx, *grad_outs):
                self.parents.append(name)
                return grad_outs

        return PushState.apply

    def create_backwards_pop(self, name):
        """
        Creates a function to be used in the backward pass of the model.

        :param name: name of the module from which it was called. This function is used to keep track of the nested
                     module hierarchy when computing the operations count.
        :return: function applies the gradient of the output to the input by calling the backward()
        """

        class PopState(Function):
            @staticmethod
            def forward(ctx, *args):
                args = tree_map(lambda x: x.clone() if isinstance(x, Tensor) else x, args)
                if len(args) == 1:
                    return args[0]
                return args

            @staticmethod
            def backward(ctx, *grad_outs):
                assert self.parents[-1] == name
                self.parents.pop()
                return grad_outs

        return PopState.apply

    def __enter__(self):
        """
        This method is called when the with block for the OperationsCounterMode object is entered. It clears the
        ops_count dictionary and calls the parent __enter__() method.
        """
        self.operations_count.clear()
        super().__enter__()
        return self

    def __exit__(self, *args):
        """
        This method is called when the with block for the OperationsCounterMode object is exited. It sorts the ops_count
        dictionary in descending order by the number of operations counted and calls the parent __exit__() method.
        """
        for mod in self.operations_count.keys():
            operations_count_mod_items = self.operations_count[mod].items()
            self.operations_count[mod] = dict(sorted(operations_count_mod_items, key=itemgetter(1), reverse=True))
        super().__exit__(*args)

    @property
    def total_operations(self):
        """
        total number of operations counted in the forward pass of the model. It does this by summing the number of
        operations for each module in the ops_count dictionary.
        """
        assert self.total_main_operation is not None, "No operations were computed to count."
        return self.total_main_operation

    def __torch_dispatch__(self, func, types, args=(), kwargs=None):
        """
        This method is called when the OperationsCounterMode object is used as a dispatch mode in a PyTorch function
        call. It updates the ops_count dictionary with the number of operations for the given function call. It also
        counts operations using the PyTorch profiler if the function call involves GPU computations.

        :param func:  The PyTorch function being called
        :param types: Tuple of argument types
        :param args: Tuple of input arguments
        :param kwargs: The dictionary of keyword argument
        :return: The output of the PyTorch function call
        """
        kwargs = kwargs if kwargs else {}
        operation_count = 0
        out = func(*args, **kwargs)
        func_packet = func._overloadpacket
        for par in self.parents:
            self.operations_count[par][func_packet] += operation_count
        if operation_count == 0:
            if func_packet in operations_mapping:
                operation_count = operations_mapping[func_packet](args, normalize_tuple(out))
                for par in self.parents:
                    self.operations_count[par][func_packet] += operation_count
            else:
                if str(func_packet) not in aten_ignore:
                    warn(f"operation `{func_packet}` is not considered in counting OPs.")

        self.total_main_operation = sum(self.operations_count[MAIN_NAME].values())

        return out



from math import prod
from numbers import Number
from typing import Any, List

from torch import ops

from torch_operation_counter.utils import conv_ops_count, transpose_shape


def basic_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    num_operations = prod(inputs[0].shape)
    return num_operations


def arange_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    num_operations = len(outputs[0])
    return num_operations


def scatter_add_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    dim = inputs[1]
    index = inputs[2]
    src = inputs[3]
    assert index.dim() == src.dim(), "index and src must have the same number of dimensions"
    num_operations = 0
    if dim < 0:
        dim += src.dim()
    for d in range(src.dim()):
        if d == dim:
            unique_indices, counts = index.select(dim, d).view(-1).unique(return_counts=True)
            num_operations += counts.sum().item()
        else:
            num_operations += src[d].numel()
    return num_operations


def scatter_reduce_ops(inputs: List[Any], outputs: List[Any]) -> int:
    reduce = inputs[4].lower()

    if reduce in ["sum", "prod", "amax", "amin"]:
        num_operations = scatter_add_ops(inputs, outputs)
    elif reduce == "mean":
        num_operations = 2 * scatter_add_ops(inputs, outputs)
    else:
        raise ValueError(f"Unknown reduce operation: {reduce}")
    return num_operations


def matmul_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    input_shapes = [v.shape for v in inputs]
    num_operations = prod(input_shapes[0]) * input_shapes[-1][-1]
    return num_operations


def addmm_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    input_shapes = [v.shape for v in inputs[1:3]]
    batch_size, input_dim = input_shapes[0]
    output_dim = input_shapes[1][1]
    num_operations = batch_size * input_dim * output_dim
    return num_operations


def bmm_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    input_shapes = [v.shape for v in inputs]
    n, c, t = input_shapes[0]
    d = input_shapes[-1][-1]
    num_operations = n * c * t * d
    return num_operations


def relu_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    num_operations = 2 * prod(inputs[0].shape)  # also count the comparison
    return num_operations


def leaky_relu_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    num_operations = 4 * prod(inputs[0].shape)  # also count the comparison
    return num_operations


def softmax_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    num_operations = 5 * prod(inputs[0].shape)
    return num_operations


def log_softmax_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    num_operations = 6 * prod(inputs[0].shape)
    return num_operations


def mean_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    dims = inputs[1]
    num_operations = prod([inputs[0].size(dim) for dim in dims]) + len(dims)
    return num_operations


def convolution_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    x, w = inputs[:2]
    x_shape, w_shape, out_shape = (x.shape, w.shape, outputs[0].shape)
    transposed = inputs[6]
    return conv_ops_count(x_shape, w_shape, out_shape, transposed=transposed)


def native_batch_norm_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    num_operations = 2 * prod(inputs[0].shape)
    # count the affine `running_mean` and `running_var` operation
    if inputs[1] is not None and inputs[2] is not None:
        num_operations = 2 * num_operations
    return num_operations


def convolution_backward_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    grad_out_shape, x_shape, w_shape = [i.shape for i in inputs[:3]]
    output_mask = inputs[-1]
    fwd_transposed = inputs[7]
    ops_count = 0
    if output_mask[0]:
        grad_input_shape = outputs[0].shape
        ops_count += conv_ops_count(grad_out_shape, w_shape, grad_input_shape, not fwd_transposed)
    if output_mask[1]:
        grad_weight_shape = outputs[1].shape
        ops_count += conv_ops_count(transpose_shape(x_shape), grad_out_shape, grad_weight_shape, fwd_transposed)
    return ops_count


def max_pool2d_with_indices_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    input_tensor = inputs[0]
    kernel_size = inputs[1] if len(inputs) > 1 else (2, 2)
    stride = inputs[2] if len(inputs) > 2 else (2, 2)
    padding = inputs[3] if len(inputs) > 3 else (0, 0)
    # Calculate output dimensions
    nC = input_tensor.shape[1]
    nH = (input_tensor.shape[2] - kernel_size[0] + 2 * padding[0]) // stride[0] + 1
    nW = (input_tensor.shape[3] - kernel_size[1] + 2 * padding[1]) // stride[1] + 1
    # Number of comparisons per window
    comparisons_per_window = prod(kernel_size) - 1
    # Total operations are comparisons across all output elements
    num_operations = comparisons_per_window * nC * nH * nW
    return num_operations


def avg_pool2d_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    input_tensor = inputs[0]
    kernel_size = inputs[1] if len(inputs) > 1 else (2, 2)
    stride = inputs[2] if len(inputs) > 2 else (2, 2)
    padding = inputs[3] if len(inputs) > 3 else (0, 0)
    # Calculate output dimensions
    nC = input_tensor.shape[1]
    nH = (input_tensor.shape[2] - kernel_size[0] + 2 * padding[0]) // stride[0] + 1
    nW = (input_tensor.shape[3] - kernel_size[1] + 2 * padding[1]) // stride[1] + 1
    comparisons_per_window = (prod(kernel_size) - 1) * nC
    num_operations = comparisons_per_window * nC * nH * nW
    return num_operations


def constant_pad_nd_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    num_operations = len(inputs[1]) * prod(inputs[0].shape)
    return num_operations


def max_pool2d_with_indices_backward_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    grad_tensor = inputs[0]  # Assuming this is the gradient tensor
    # Each element in the gradient tensor corresponds to an operation
    num_operations = prod(grad_tensor.shape)
    return num_operations


def native_batch_norm_backward_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    num_operations = 2 * prod(inputs[0].shape)
    return num_operations


def threshold_backward_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    num_operations = 2 * prod(inputs[0].shape)
    return num_operations


def gather_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    index_tensor = inputs[2]  # Assuming this is the index tensor
    num_operations = prod(index_tensor.shape)
    return num_operations


def index_add_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    source_tensor = inputs[3]  # Assuming this is the source tensor
    num_operations = 2 * prod(source_tensor.shape)
    return num_operations


def new_zeros_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    output_tensor = outputs[0]
    # Each element involves a write operation
    num_operations = prod(output_tensor.shape)
    return num_operations


def masked_fill_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    mask_tensor = inputs[1]
    # Count true elements in the mask as operations (read mask, write value)
    num_operations = (mask_tensor == True).sum().item()
    return num_operations


def full_ops(inputs: List[Any], outputs: List[Any]) -> Number:
    output_tensor = outputs[0]
    num_operations = prod(output_tensor.shape)  # Each element set once
    return num_operations


operations_mapping = {
    ops.aten.min: basic_ops,
    ops.aten.max: basic_ops,
    ops.aten.sort: basic_ops,
    ops.aten.eq: basic_ops,
    ops.aten.ne: basic_ops,
    ops.aten.add_: basic_ops,
    ops.aten.add: basic_ops,
    ops.aten.sub: basic_ops,
    ops.aten.mul: basic_ops,
    ops.aten.mul_: basic_ops,
    ops.aten.pow_: basic_ops,
    ops.aten.exp: basic_ops,
    ops.aten.sum: basic_ops,
    ops.aten.div: basic_ops,
    ops.aten.div_: basic_ops,
    ops.aten.bitwise_not: basic_ops,
    # None-basic operations
    ops.aten.scatter_add: scatter_add_ops,
    ops.aten.scatter_add_: scatter_add_ops,
    ops.aten.scatter_reduce: scatter_reduce_ops,
    ops.aten.scatter_reduce_: scatter_reduce_ops,
    ops.aten.arange: arange_ops,
    ops.aten.mean: mean_ops,
    ops.aten.mm: matmul_ops,
    ops.aten.matmul: matmul_ops,
    ops.aten.addmm: addmm_ops,
    ops.aten.bmm: bmm_ops,
    ops.aten.relu: relu_ops,
    ops.aten.relu_: relu_ops,
    ops.aten.leaky_relu: leaky_relu_ops,
    ops.aten.leaky_relu_: leaky_relu_ops,
    ops.aten.elu: leaky_relu_ops,
    ops.aten.elu_: leaky_relu_ops,
    ops.aten._softmax: softmax_ops,
    ops.aten._log_softmax: log_softmax_ops,
    ops.aten.native_batch_norm: native_batch_norm_ops,
    ops.aten.convolution: convolution_ops,
    ops.aten._convolution: convolution_ops,
    ops.aten.convolution_backward: convolution_backward_ops,
    ops.aten.max_pool2d_with_indices: max_pool2d_with_indices_ops,
    ops.aten.max_pool2d_with_indices_backward: max_pool2d_with_indices_backward_ops,
    ops.aten.avg_pool2d: avg_pool2d_ops,
    ops.aten.constant_pad_nd: constant_pad_nd_ops,
    ops.aten.native_batch_norm_backward: native_batch_norm_backward_ops,
    ops.aten.threshold_backward: threshold_backward_ops,
    ops.aten.gather: gather_ops,
    ops.aten.index_add: index_add_ops,
    ops.aten.index_add_: index_add_ops,
    ops.aten.new_zeros: new_zeros_ops,
    ops.aten.masked_fill: masked_fill_ops,
    ops.aten.masked_fill_: masked_fill_ops,
    ops.aten.full: full_ops,
}



import unittest

from torch import Tensor
from torch.nn import Module

from torch_operation_counter import OperationsCounterMode


class MatrixMultiplicationModule(Module):
    def __init__(self, A: Tensor, B: Tensor):
        super().__init__()
        self.A = A
        self.B = B

    def forward(self):
        return self.A @ self.B


class TestMatrixMultiplication(unittest.TestCase):
    def test_correctness(self):
        A = Tensor([[1, 2],
                    [3, 4]])
        B = Tensor([[1, 2],
                    [3, 4]])
        module = MatrixMultiplicationModule(A, B)
        output = module()
        expected_output = A @ B
        self.assertTrue((output == expected_output).all())

    def test_num_ops(self):
        A = Tensor([[1, 2],
                    [3, 4]])
        B = Tensor([[1, 2],
                    [3, 4]])
        module = MatrixMultiplicationModule(A, B)
        # A \in R^{2x2}, B \in R^{2x2}, C = A @ B \in R^{2x2}
        # C_{ij} = \sum_{k=1}^{2} A_{ik} * B_{kj}
        # In each element of C, there are 2 multiplications.
        # There are 4 elements in C
        # Therefore, the total number of operations is 4 * 2 = 8
        with OperationsCounterMode() as ops_counter:
            module()
        self.assertEqual(ops_counter.total_operations, 8)


if __name__ == "__main__":
    unittest.main()